# -*- coding: utf-8 -*-


def sanitize_metric_name(key):
    """Ensure metrics are statsd/datadog compliant, ie use lower case and
    dot-seperated namespace.

    Examples:
        >>> # change nothing on good format
        >>> sanitize_metric_name('some.metrics.name')
        'some.metrics.name'
        >>> # otherwise:
        >>> sanitize_metric_name('SOME.weird.name')
        'some.weird.name'
        >>> sanitize_metric_name('some/weird/name')
        'some.weird.name'

    """
    return key.lower().replace('/', '.')


def make_list(value, sep=','):
    """Split strings around `sep` or return as is.

    Examples:
        >>> make_list('foo,bar')
        ['foo', 'bar']
        >>> make_list('foo,bar', sep='.')
        ['foo,bar']
        >>> make_list(3)
        3

    """
    return value.split(sep) if isinstance(value, str) else value



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals
import logging
import os

from scrapy.exceptions import NotConfigured
from datadog.api.constants import CheckStatus
import six

logger = logging.getLogger(__name__)

# define sane default behavior of the extension
DEFAULT_CUSTOM_METRICS = [
    # we usually want to monitor the quantity of data found and the performance
    'item_scraped_count',
    'response_received_count',

    # logs can be a good indicator of something going wrong
    'log_count/WARNING',
    'log_count/ERROR',
]
DEFAULT_MISSING = 0
DEFAULT_DD_PREFIX = 'spider.stats'
DEFAULT_DD_HOST_NAME = 'app.scrapinghub.com'
DEFAULT_SERVICE_CHECK = 'data.is_sourced'

MANDATORY_SETTINGS = [
    'DATADOG_API_KEY',
    'DATADOG_APP_KEY',
    'SCRAPY_PROJECT_ID',
]
CUSTOM_SETTINGS = [
    'DATADOG_HOST_NAME',
    'DATADOG_METRICS_PREFIX',
    'DATADOG_SERVICE_CHECK',
    'DATADOG_CUSTOM_METRICS',
    'DATADOG_CUSTOM_TAGS',
]

# read more: http://help.scrapinghub.com/scrapy-cloud/job-outcomes
# TODO support `close_spider_*`
DD_STATUS_MAPPING = {
    'finished': CheckStatus.OK,
    'no_reason': CheckStatus.UNKNOWN,
    'failed': CheckStatus.CRITICAL,
    'cancelled': CheckStatus.WARNING,
    'cancel_timeout': CheckStatus.CRITICAL,
    'shutdown': CheckStatus.CRITICAL,
    'banned': CheckStatus.CRITICAL,
    'memusage_exceeded': CheckStatus.CRITICAL,
    'slybot_fewitems_scraped': CheckStatus.WARNING,
    'default': CheckStatus.UNKNOWN,
}


def getbool(conf, key):
    """Emulation of scrapy.Settings;getbool.

    It allows the config to remain agnostic from Scrapy specificities. Not that
    it brings much to the table but that's really cheap to make it so.

    Examples:
        >>> getbool({}, 'foo')
        False
        >>> getbool({'foo': 'no'}, 'foo')
        False
        >>> getbool({'foo': 'yes'}, 'foo')
        True
        >>> getbool({'foo': 'True'}, 'foo')
        True

    """
    return conf.get(key) in [True, 'True', 'true', 'yes', 'y']


def validate_conf(conf):
    """Abort extension setup if we don't have all the required settings."""
    # shortcut to disable the extension - especially useful for local
    # development that doesn't require metrics
    # if conf.get()
    if getbool(conf, 'DATADOG_DISABLED'):
        logger.warning('extension explicitely disabled')
        raise NotConfigured

    for key in MANDATORY_SETTINGS:
        if key not in conf:
            logger.warning('datadog extension mandatory setting missing: {}'.format(key))
            logger.warning('scraper will go on but extension will be disabled.')
            raise NotConfigured
        elif conf[key] is None:
            logger.warning('setting {} was set to `None`.'.format(key))
            raise NotConfigured


def merge_env(settings):
    """Merge settings setup from env and/or scrapy settings.

    Note that Scrapy `settings` take precedence over env and won't be modified
    if already set.

    Also, this method doesn't return anything. Instead it mutates the
    given dict on purpose.

    """
    ALL_SETTINGS = MANDATORY_SETTINGS + CUSTOM_SETTINGS

    settings.update({
        k: v for k, v in six.iteritems(os.environ)
        if k in ALL_SETTINGS and k not in settings
    })



# -*- coding: utf-8 -*-

"""Scrapy extension sending crawl stats to Datadog.

## References

datadog api: http://docs.datadoghq.com/api/
stats api:   https://doc.scrapy.org/en/latest/topics/stats.html

## Scrapy built-in stats

    - downloader/request_bytes
    - downloader/request_count': 18,
    - downloader/request_method_count/GET': 4,
    - downloader/request_method_count/POST': 14,
    - downloader/response_bytes': 155010,
    - downloader/response_count': 18,
    - downloader/response_status_count/200': 18,
    - finish_reason': 'finished',
    - finish_time': datetime.datetime(2017, 4, 3, 9, 13, 46, 96057),
    - item_scraped_count': 4,
    - log_count/${level}': 29,
    - request_depth_max': 12,
    - response_received_count': 18,
    - scheduler/dequeued': 18,
    - scheduler/dequeued/memory': 18,
    - scheduler/enqueued': 18,
    - scheduler/enqueued/memory': 18,
    - start_time': datetime.datetime(2017, 4, 3, 9, 12, 20, 281530)}

## Extension built-in stats (always sent)

    - done: ping datadog (i.e. just send 1)
    - elapsed_time: time in second betwwen spider start and finish
    - exit_code: unix-like code to datadog to understand finish state

## Extension built-in tags

    - spider_name
    - project_id

"""

# env: scrapinghub
# role: scraping
# stack: sourcing

from __future__ import absolute_import, unicode_literals
import logging

import datadog
from datadog.api.constants import CheckStatus
from scrapy import signals

from scrapydatadog import utils
# prevent Scrapy `settings` overwrite
from scrapydatadog import settings as ext_settings

logger = logging.getLogger(__name__)


class DatadogExtension(object):

    def __init__(self, settings, stats):
        # properly identify the job
        self.project_id = settings['SCRAPY_PROJECT_ID']

        self.dd_api_key = settings['DATADOG_API_KEY']
        self.dd_app_key = settings['DATADOG_APP_KEY']
        self.dd_host_name = settings.get('DATADOG_HOST_NAME', ext_settings.DEFAULT_DD_HOST_NAME)
        self.dd_metric_prefix = settings.get('DATADOG_METRICS_PREFIX',
                                             ext_settings.DEFAULT_DD_PREFIX)
        self.dd_service_check = settings.get('DATADOG_SERVICE_CHECK',
                                             ext_settings.DEFAULT_SERVICE_CHECK)

        # extend default behaviors
        self.to_collect = utils.make_list(settings.get('DATADOG_CUSTOM_METRICS',
                                          ext_settings.DEFAULT_CUSTOM_METRICS))
        self.custom_tags = utils.make_list(settings.get('DATADOG_CUSTOM_TAGS', []))

        # allow the rest of the extension to access stats logic/data
        self.stats = stats

    @classmethod
    def from_crawler(cls, crawler):
        ext_settings.merge_env(crawler.settings)
        ext_settings.validate_conf(crawler.settings)

        ext = cls(crawler.settings, crawler.stats)

        crawler.signals.connect(ext.spider_closed, signal=signals.spider_closed)

        return ext

    def build_tags(self, spider):
        """Aggregate and format spider tags.

        The extension merges tags from 3 different places:

            - built-in: always good to segment by project and spider
            - settings: a common place for generic, static or poecjt wide tags
            - `spider.tags` to let spiders dynamically set them after init

        """
        dynamic_tags = getattr(spider, 'tags', [])

        return ['project:{}'.format(self.project_id),
                'spider_name:{}'.format(spider.name),
                ] + self.custom_tags + dynamic_tags

    def _metrics_fmt(self, key):
        sane_key = utils.sanitize_metric_name(key)
        return '{}.{}'.format(self.dd_metric_prefix, sane_key)

    def commit(self, metrics):
        res = datadog.api.Metric.send(metrics)
        logger.debug('API call result: {}'.format(res))

    def publish_status(self, spider_name, finish_reason, tags):
        status = ext_settings.DD_STATUS_MAPPING.get(finish_reason, CheckStatus.UNKNOWN)
        msg = 'Spider {} ended because: {}'.format(spider_name, finish_reason)

        logger.info(msg)
        datadog.api.ServiceCheck.check(check=self.dd_service_check,
                                       status=status,
                                       message=msg,
                                       host=ext_settings.DEFAULT_DD_HOST_NAME,
                                       tags=tags)

    def spider_closed(self, spider):
        # initialize API client
        logger.info('connecting to datadgo API')
        options = {
            'host_name': self.dd_host_name,
            'api_key': self.dd_api_key,
            'app_key': self.dd_app_key,
        }
        datadog.initialize(**options)

        # Fetch scrapy stats
        job_stats = self.stats.get_stats(spider)
        logger.debug('ScrapyStats from crawler: {}'.format(job_stats))

        # Build metrics list of dict to send to Datadog API
        tags = self.build_tags(spider)

        def collect(key, value):
            """Format intuitive key=value into datadog format.

            ref: http://datadogpy.readthedocs.io/en/latest/#datadog.api.Metric.send

            """
            return {
                'metric': self._metrics_fmt(key),
                'points': value,
                'tags': tags,
                'host': self.dd_host_name,
            }

        metrics = []

        if 'finish_time' in list(job_stats.keys()) and 'start_time' in list(job_stats.keys()):
            elapsed_time = job_stats['finish_time'] - job_stats['start_time']
            metrics.append(collect('elapsed_time', elapsed_time.seconds))

        # TODO allow regex matching
        for key in self.to_collect:
            metrics.append(collect(key, job_stats.get(key, ext_settings.DEFAULT_MISSING)))

        logger.info('collected metrics, publishing to datadog')
        self.commit(metrics)

        # notify success/failure using DD checks
        self.publish_status(spider.name, job_stats.get('finish_reason'), tags)



# -*- coding: utf-8; -*-

from __future__ import absolute_import, unicode_literals
import os
import unittest

from mock import Mock
from scrapy.exceptions import NotConfigured
from scrapy.settings import Settings

from scrapydatadog.extension import DatadogExtension
from scrapydatadog.settings import merge_env


class TestDatadogConfig(unittest.TestCase):

    def test_unconfigured_init(self):
        crawler = Mock()
        # don't include `DATADOG_API_KEY` and make sure it's not in the env
        os.environ.pop('DATADOG_API_KEY', None)
        crawler.settings = Settings({'DATADOG_APP_KEY': 'azertyuiop123456789'})

        with self.assertRaises(NotConfigured):
            self.extension = DatadogExtension.from_crawler(crawler)

    def test_configured_init(self):
        dd_api_key = 'azertyuiop123456789'
        dd_app_key = 'azertyuiop123456789'
        scrapy_id = '000'

        crawler = Mock()
        crawler.settings = Settings({
            'DATADOG_API_KEY': dd_api_key,
            'DATADOG_APP_KEY': dd_app_key,
            'SCRAPY_PROJECT_ID': scrapy_id,
        })

        raised = None
        try:
            DatadogExtension.from_crawler(crawler)
        except Exception as e:  # noqa
            raised = e
        self.assertIsNone(raised, 'Exception raised: {}'.format(raised))

    def test_merging_environ_in_settings(self):
        # add something that shouldn't be read
        os.environ['HELLO'] = 'world'
        # add a custom setting
        os.environ['DATADOG_HOST_NAME'] = 'localhost'
        # and a mandatory setting
        os.environ['DATADOG_API_KEY'] = 'xxxxx'
        # and a normal init + something arbitrary that should stay there
        settings = Settings({'FOO': 'bar'})

        merge_env(settings)

        self.assertEqual(settings['FOO'], 'bar')
        self.assertEqual(settings['DATADOG_HOST_NAME'], 'localhost')
        self.assertEqual(settings['DATADOG_API_KEY'], 'xxxxx')
        self.assertIsNone(settings.get('HELLO'))



from setuptools import find_packages, setup


setup(
    name='scrapy-datadog',
    description='Scrapy extension to send scrapy stats to Datadog.',
    author='Kpler engineering',
    author_email='engineering@kpler.com',
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    packages=find_packages(),
    url='http://github.com/kpler/scrapy-datadog-extension',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
    ],
    install_requires = [
        'scrapy',
        'datadog',
    ],
    extras_require={
        "test": [
            'coverage==4.5',
            'flake8==3.5.0',
            'ipdb==0.10.3',
            'mock==2.0.0',
            'nose==1.3.7',
            'nose-cov==1.6',
            'nose-timer==0.7.1',
            'tox==3.0.0',
            'tox-pyenv==1.1.0'
        ]
    }
)
